/*
 *	This is a part of adsbox - ADS-B decoder software
 * 	(c) atty 2011-12 romulmsg@mail.ru
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "../sqlite3/sqlite3.h"
#include "misc.h"
#include "db.h"

#define CPR_EXPIRE 15 // seconds to odd/even CPR pair expire
#define DROP_AIRCRAFT 60 // seconds to track expire
extern sqlite3 *db;
extern double homelat;
extern double homelon;

extern void azimuth(double *lat, double *lon, double *lat1, double *lon1, double *a);
extern double dist(double lat, double lon, double lat1, double lon1);
extern double round(double value);

char *stmts[] = {
                    "CREATE TABLE FLIGHTS (ICAO INTEGER PRIMARY KEY ASC, \
                    LAT0 INTEGER DEFAULT 0, LON0 INTEGER DEFAULT 0, LAT1 INTEGER DEFAULT 0, \
                    LON1 INTEGER DEFAULT 0, LAT REAL DEFAULT 0, LON REAL DEFAULT 0, \
                    ALT INTEGER DEFAULT 0, CALLSIGN TEXT DEFAULT '', SPEED REAL DEFAULT 0, VSPEED DEFAULT 0, \
                    HEADING REAL DEFAULT 0, SQUAWK INTEGER DEFAULT 0, SEEN_TIME INTEGER  \
                    DEFAULT ((julianday('now') - 2440587.5)*86400), PARITY INTEGER DEFAULT 0)\0",
                    "CREATE INDEX IX_FLIGHTS ON FLIGHTS (ICAO)\0",
                    "CREATE TRIGGER ON_DELETE_FLIGHTS AFTER DELETE ON FLIGHTS BEGIN \
                    DELETE FROM TRACKS WHERE ICAO=OLD.ICAO; END;\0",
                    "CREATE TRIGGER ON_UPDATE_FLIGHTS AFTER UPDATE ON FLIGHTS BEGIN \
                    UPDATE FLIGHTS SET SEEN_TIME = (julianday('now') - 2440587.5)*86400 \
                    WHERE ICAO=OLD.ICAO; END;\0",
                    "CREATE TABLE RADAR_RANGE (AZIMUTH REAL DEFAULT 0, \
                    LAT DEFAULT 0, LON DEFAULT 0, DIST REAL DEFAULT 0)\0",
                    "CREATE TABLE TRACKS (ICAO INTEGER, \
                    LAT REAL DEFAULT 0, LON REAL DEFAULT 0, ALT INTEGER DEFAULT 0, \
                    SEEN_TIME INTEGER DEFAULT ((julianday('now') - 2440587.5)*86400) )\0",
                    "CREATE INDEX IX_TRACKS ON TRACKS (ICAO)\0",
                    NULL
                };

int db_update_icao(unsigned int *icao) {
    int rc;
    char *zErrMsg = 0;
    char *sql = malloc(256);

    //sqlite3_mutex_enter(db_mux);
    sprintf(
        sql,
        "UPDATE FLIGHTS SET ICAO=%i WHERE ICAO=%i\0", *icao, *icao);
    rc = sqlite3_exec(db, sql, NULL, 0, &zErrMsg);
    if (rc != SQLITE_OK) {
        DEBUG("SQLite error: %s\n", zErrMsg);
        sqlite3_free(zErrMsg);
        free(sql);
        //sqlite3_mutex_leave(db_mux);
        return 1;
    }

    if (!sqlite3_changes(db)) {
        sprintf(
            sql,
            "INSERT INTO FLIGHTS (ICAO) VALUES (%i)\0", *icao);
        rc = sqlite3_exec(db, sql, NULL, 0, &zErrMsg);
        if (rc != SQLITE_OK) {
            DEBUG("SQLite error: %s\n", zErrMsg);
            sqlite3_free(zErrMsg);
            free(sql);
            //sqlite3_mutex_leave(db_mux);
            return 1;
        }
    }

    free(sql);
    //sqlite3_mutex_leave(db_mux);
    return 0;
}

int db_update_position(unsigned int *icao, double *lat, double *lon) {
    int rc;
    char *zErrMsg = 0;
    char *sql = malloc(256);
    double a, d;

    //sqlite3_mutex_enter(db_mux);
    /*
     *  Update record in flights table
     */
    sprintf(
        sql,
        "UPDATE FLIGHTS SET LAT=%f, LON=%f WHERE ICAO=%i\0",
        *lat, *lon, *icao);
    rc = sqlite3_exec(db, sql, NULL, 0, &zErrMsg);
    if (rc != SQLITE_OK) {
        DEBUG("SQLite error: %s in %s\n", zErrMsg, sql);
        sqlite3_free(zErrMsg);
        free(sql);
        //sqlite3_mutex_leave(db_mux);
        return 1;
    }

    /*
     * If nothing was updated create the record
     */
    if (!sqlite3_changes(db)) {
        sprintf(
            sql,
            "INSERT INTO FLIGHTS (ICAO, LAT0, LON0, LAT1, LON1, \
            LAT, LON, ALT, CALLSIGN, SPEED, HEADING, SQUAWK) \
            VALUES (%i,0,0,0,0,%f,%f,0,'',0,0,0)\0",
            *icao, *lat, *lon);
        rc = sqlite3_exec(db, sql, NULL, 0, &zErrMsg);
        if (rc != SQLITE_OK) {
            DEBUG("SQLite error: %s in %s\n", zErrMsg, sql);
            sqlite3_free(zErrMsg);
            free(sql);
            //sqlite3_mutex_leave(db_mux);
            return 1;
        }
    }
    /*
     *  Delete expired flight
     */
    sprintf(
        sql,
        "DELETE FROM FLIGHTS \
        WHERE (julianday('now') - 2440587.5)*86400 - SEEN_TIME > %i\0",
        DROP_AIRCRAFT);
    rc = sqlite3_exec(db, sql, NULL, 0, &zErrMsg);
    if (rc != SQLITE_OK) {
        DEBUG("SQLite error: %s in %s\n", zErrMsg, sql);
        sqlite3_free(zErrMsg);
        free(sql);
        //sqlite3_mutex_leave(db_mux);
        return 1;
    }

    /*
     * Record the track point
     */
    sprintf(
        sql,
        "INSERT INTO TRACKS (ICAO, LAT, LON, ALT) \
        SELECT %i, %f, %f, ALT FROM FLIGHTS \
        WHERE FLIGHTS.ICAO=%i AND FLIGHTS.ALT>0\0",
        *icao, *lat, *lon, *icao);
    rc = sqlite3_exec(db, sql, NULL, 0, &zErrMsg);
    if (rc != SQLITE_OK) {
        DEBUG("SQLite error: %s in %s\n", zErrMsg, sql);
        sqlite3_free(zErrMsg);
        free(sql);
        //sqlite3_mutex_leave(db_mux);
        return 1;
    }

    /*
     *  Update radar coverage area
     */
    if (homelat != 0 && homelon != 0) {
        azimuth(&homelat, &homelon, lat, lon, &a);
        d = dist(homelat, homelon, *lat, *lon);
        sprintf(
            sql,
            "UPDATE RADAR_RANGE SET \
            LAT=%f, LON=%f, DIST=%f WHERE ABS(AZIMUTH - %f) < 1 AND DIST<%f\0",
            *lat, *lon, d, a, d);
        rc = sqlite3_exec(db, sql, NULL, 0, &zErrMsg);
        if (rc != SQLITE_OK) {
            DEBUG("SQLite error: %s in %s\n", zErrMsg, sql);
            sqlite3_free(zErrMsg);
            free(sql);
            //sqlite3_mutex_leave(db_mux);
            return 1;
        }
    }

    free(sql);
    //sqlite3_mutex_leave(db_mux);
    return 0;
}

int db_update_alt(unsigned int *icao, int *alt) {
    int rc;
    char *zErrMsg = 0;
    char *sql = malloc(256);
    int rslt = 0;

    //sqlite3_mutex_enter(db_mux);

    sprintf(sql, "UPDATE FLIGHTS SET ALT=ROUND(0.3048 * %i) WHERE ICAO=%i\0",
            *alt, *icao);
    rc = sqlite3_exec(db, sql, NULL, 0, &zErrMsg);
    if (rc != SQLITE_OK) {
        DEBUG("SQLite error: %s\n", zErrMsg);
        sqlite3_free(zErrMsg);
        free(sql);
        //sqlite3_mutex_leave(db_mux);
        return -1;
    }

    rslt = sqlite3_changes(db);

    free(sql);
    //sqlite3_mutex_leave(db_mux);
    return rslt;
}

int db_update_speed_heading(unsigned int *icao, float *speed, float *heading,
                            int *vr) {
    int rc;
    char *zErrMsg = 0;
    char *sql = malloc(256);

    //sqlite3_mutex_enter(db_mux);
    sprintf(
        sql,
        "UPDATE FLIGHTS SET SPEED=%f, VSPEED=%i, HEADING=%f WHERE ICAO=%i\0",
        *speed, *vr, *heading, *icao);
    rc = sqlite3_exec(db, sql, NULL, 0, &zErrMsg);
    if (rc != SQLITE_OK) {
        DEBUG("SQLite error: %s\n", zErrMsg);
        sqlite3_free(zErrMsg);
        free(sql);
        //sqlite3_mutex_leave(db_mux);
        return 1;
    }

    if (!sqlite3_changes(db)) {
        sprintf(
            sql,
            "INSERT INTO FLIGHTS (ICAO, SPEED, VSPEED, HEADING) \
            VALUES (%i,%f,%i,%f)\0",
            *icao, *speed, *vr, *heading);
        rc = sqlite3_exec(db, sql, NULL, 0, &zErrMsg);
        if (rc != SQLITE_OK) {
            DEBUG("SQLite error: %s\n", zErrMsg);
            sqlite3_free(zErrMsg);
            free(sql);
            //sqlite3_mutex_leave(db_mux);
            return 1;
        }
    }

    free(sql);
    //sqlite3_mutex_leave(db_mux);
    return 0;
}

int db_update_squawk(unsigned int *icao, unsigned int *squawk) {
    int rc;
    char *zErrMsg = 0;
    char *sql = malloc(256);
    int rslt = 0;

    //sqlite3_mutex_enter(db_mux);
    sprintf(sql, "UPDATE FLIGHTS SET SQUAWK=%i WHERE ICAO=%i\0", *squawk, *icao);
    rc = sqlite3_exec(db, sql, NULL, 0, &zErrMsg);
    if (rc != SQLITE_OK) {
        DEBUG("SQLite error: %s\n", zErrMsg);
        sqlite3_free(zErrMsg);
        free(sql);
        //sqlite3_mutex_leave(db_mux);
        return -1;
    }

    rslt = sqlite3_changes(db);

    free(sql);
    //sqlite3_mutex_leave(db_mux);
    return rslt;
}

int db_update_callsign(unsigned int *icao, char *callsign) {
    int rc;
    char *zErrMsg = 0;
    char *sql = malloc(256);

    //sqlite3_mutex_enter(db_mux);
    sprintf(sql, "UPDATE FLIGHTS SET CALLSIGN=TRIM('%s') WHERE ICAO=%i\0",
            (char *) callsign, *icao);
    rc = sqlite3_exec(db, sql, NULL, 0, &zErrMsg);
    if (rc != SQLITE_OK) {
        DEBUG("SQLite error: %s\n", zErrMsg);
        sqlite3_free(zErrMsg);
        free(sql);
        return 1;
        //sqlite3_mutex_leave(db_mux);
    }

    if (!sqlite3_changes(db)) {
        sprintf(
            sql,
            "INSERT INTO FLIGHTS (ICAO, LAT0, LON0, LAT1, LON1, LAT, LON, ALT, CALLSIGN, SPEED, HEADING, SQUAWK) \
            VALUES (%i,0,0,0,0,0,0,0,TRIM('%s'),0,0,0)\0",
            *icao, (char *) callsign);
        rc = sqlite3_exec(db, sql, NULL, 0, &zErrMsg);
        if (rc != SQLITE_OK) {
            DEBUG("SQLite error: %s\n", zErrMsg);
            sqlite3_free(zErrMsg);
            free(sql);
            //sqlite3_mutex_leave(db_mux);
            return 1;
        }
    }

    free(sql);
    //sqlite3_mutex_leave(db_mux);
    return 0;
}

cpr_info db_update_cpr(unsigned int *icao, unsigned char *f, int *lat, int *lon) {
    int rc;
    char *zErrMsg = 0;
    char *sql = malloc(256);
    sqlite3_stmt *stmt;
    cpr_info ci;
    long int seen_time;
    time_t t = time(NULL);

    ci.lat0 = 0;
    ci.lon0 = 0;
    ci.lat1 = 0;
    ci.lon1 = 0;
    ci.parity = 0;
    ci.lat = 0;
    ci.lon = 0;
    seen_time = t;

    //sqlite3_mutex_enter(db_mux);

    sprintf(
        sql,
        "SELECT LAT0, LON0, LAT1, LON1, PARITY, SEEN_TIME, LAT, LON \
        FROM FLIGHTS WHERE ICAO=%i\0",
        *icao);
    rc = sqlite3_prepare(db, sql, -1, &stmt, 0);
    if (rc != SQLITE_OK) {
        DEBUG("SQLite error: could not prepare.");
        free(sql);
        sqlite3_finalize(stmt);
        //sqlite3_mutex_leave(db_mux);
        return ci;
    }
    while (sqlite3_step(stmt) == SQLITE_ROW) {
        ci.lat0 = sqlite3_column_int(stmt, 0);
        ci.lon0 = sqlite3_column_int(stmt, 1);
        ci.lat1 = sqlite3_column_int(stmt, 2);
        ci.lon1 = sqlite3_column_int(stmt, 3);
        ci.parity = sqlite3_column_int(stmt, 4);
        seen_time = sqlite3_column_int(stmt, 5);
        ci.lat = sqlite3_column_double(stmt, 6);
        ci.lon = sqlite3_column_double(stmt, 7);
    }

    if (*f == 1) {
        ci.lat1 = *lat;
        ci.lon1 = *lon;
        if (t - seen_time > CPR_EXPIRE) {
            ci.lat0 = 0;
            ci.lon0 = 0;
        }
    } else {
        ci.lat0 = *lat;
        ci.lon0 = *lon;
        if (t - seen_time > CPR_EXPIRE) {
            ci.lat1 = 0;
            ci.lon1 = 0;
        }
    }

    sprintf(
        sql,
        "UPDATE FLIGHTS SET LAT0=%i, LON0=%i, LAT1=%i, LON1=%i, \
        PARITY=%i WHERE ICAO=%i\0",
        ci.lat0, ci.lon0, ci.lat1, ci.lon1, *f, *icao);

    rc = sqlite3_exec(db, sql, NULL, 0, &zErrMsg);
    if (rc != SQLITE_OK) {
        DEBUG("SQLite error: %s\n", zErrMsg);
        sqlite3_free(zErrMsg);
        free(sql);
        //sqlite3_mutex_leave(db_mux);
        return ci;
    }

    if (!sqlite3_changes(db)) {
        sprintf(
            sql,
            "INSERT INTO FLIGHTS (ICAO, LAT0, LON0, LAT1, LON1, PARITY) VALUES \
            (%i, %i, %i, %i, %i, %i)\0",
            *icao, ci.lat0, ci.lon0, ci.lat1, ci.lon1, *f);

        rc = sqlite3_exec(db, sql, NULL, 0, &zErrMsg);
        if (rc != SQLITE_OK) {
            DEBUG("SQLite error: %s\n", zErrMsg);
            sqlite3_free(zErrMsg);
            free(sql);
            //sqlite3_mutex_leave(db_mux);
            return ci;
        }
    }

    sqlite3_finalize(stmt);
    free(sql);
    //sqlite3_mutex_leave(db_mux);
    return ci;
}

int db_init(sqlite3 *db) {
    int rc;
    char *zErrMsg = 0;
    char *sql = malloc(128);
    int i;

    DEBUG("Init database.\n");

    char **stmt = stmts;
    while (*stmt) {
        rc = sqlite3_exec(db, *stmt, NULL, 0, &zErrMsg);
        if (rc != SQLITE_OK) {
            DEBUG("SQLite error: %s\n", zErrMsg);
            sqlite3_free(zErrMsg);
            free(sql);
            return 1;
        }
        stmt++;
    }

    /*
     *  Fill with initial values
     */
    for (i = 0; i < 360; i = i + 2) {
        sprintf(
            sql,
            "INSERT INTO RADAR_RANGE(AZIMUTH, LAT, LON) VALUES(%i, %f, %f)",
            i, homelat, homelon);
        rc = sqlite3_exec(db, sql, NULL, 0, &zErrMsg);
        if (rc != SQLITE_OK) {
            DEBUG("SQLite error: %s\n", zErrMsg);
            sqlite3_free(zErrMsg);
            free(sql);
            return 1;
        }
    }

    free(sql);
    return 0;
}
